import numpy as np
import pytest
from c3.c3objs import Quantity
from c3.generator.generator import Generator
from c3.generator.devices import (
    LO,
    AWG,
    Mixer,
    DigitalToAnalog,
    VoltsToHertz,
)
from c3.signal.pulse import Envelope, Carrier
from c3.libraries.envelopes import pwc
from c3.signal.gates import Instruction

sim_res = 100e9  # Resolution for numerical simulation
awg_res = 2e9  # Realistic, limited resolution of an AWG

lo = LO(name="lo", resolution=sim_res, outputs=1)
awg = AWG(name="awg", resolution=awg_res, outputs=1)

dac = DigitalToAnalog(name="dac", resolution=sim_res, inputs=1, outputs=1)
mixer = Mixer(name="mixer", inputs=2, outputs=1)
v_to_hz = VoltsToHertz(
    name="v_to_hz",
    V_to_Hz=Quantity(value=1e9, min_val=0.9e9, max_val=1.1e9, unit="Hz/V"),
    inputs=1,
    outputs=1,
)

generator = Generator(
    devices={
        "LO": lo,
        "AWG": awg,
        "DigitalToAnalog": dac,
        "Mixer": mixer,
        "VoltsToHertz": v_to_hz,
    },
    chains={
        "d1": {
            "LO": [],
            "AWG": [],
            "DigitalToAnalog": ["AWG"],
            "Mixer": ["LO", "DigitalToAnalog"],
            "VoltsToHertz": ["Mixer"],
        },
    },
)

t_final = 7e-9  # Time for single qubit gates
slices = int(t_final * awg_res)

pwc_params = {
    "inphase": Quantity(value=np.random.randn(slices), unit="V"),
    "quadrature": Quantity(value=np.random.randn(slices), unit="V"),
    "amp": Quantity(value=1.0, unit="V"),
    "xy_angle": Quantity(
        value=0.0, min_val=-0.5 * np.pi, max_val=2.5 * np.pi, unit="rad"
    ),
    "freq_offset": Quantity(value=0, min_val=-5 * 1e6, max_val=5 * 1e6, unit="Hz 2pi"),
}

pwc_env_single = Envelope(
    name="gauss",
    desc="Gaussian comp for single-qubit gates",
    params=pwc_params,
    shape=pwc,
)

lo_freq_q1 = 5e9
carrier_parameters = {
    "freq": Quantity(value=lo_freq_q1, min_val=4.5e9, max_val=6e9, unit="Hz 2pi"),
    "framechange": Quantity(value=0.0, min_val=-np.pi, max_val=3 * np.pi, unit="rad"),
}
carr = Carrier(
    name="carrier", desc="Frequency of the local oscillator", params=carrier_parameters
)

pulse_gate = Instruction(name="pulse", t_start=0.0, t_end=t_final, channels=["d1"])
pulse_gate.add_component(pwc_env_single, "d1")
pulse_gate.add_component(carr, "d1")


@pytest.mark.unit
def test_generation() -> None:
    signal = generator.generate_signals(pulse_gate)
    assert signal["d1"]["values"] is not None



"""Library of Hamiltonian functions."""

import numpy as np

hamiltonians = dict()


def hamiltonian_reg_deco(func):
    """
    Decorator for making registry of functions
    """
    hamiltonians[str(func.__name__)] = func
    return func


@hamiltonian_reg_deco
def resonator(a):
    """
    Harmonic oscillator hamiltonian given the annihilation operator.

    Parameters
    ----------
    a : Tensor
        Annihilator.

    Returns
    -------
    Tensor
        Number operator.

    """
    a_dag = a.T.conj()
    return np.matmul(a_dag, a)


@hamiltonian_reg_deco
def duffing(a):
    """
    Anharmonic part of the duffing oscillator.

    Parameters
    ----------
    a : Tensor
        Annihilator.

    Returns
    -------
    Tensor
        Number operator.

    """
    a_dag = a.T.conj()
    n = np.matmul(a_dag, a)
    return 1 / 2 * np.matmul(n - np.eye(int(n.shape[0]), dtype=np.complex128), n)


@hamiltonian_reg_deco
def third_order(a):
    """

    Parameters
    ----------
    a : Tensor
        Annihilator.
    Returns
    -------
    Tensor
        Number operator.
    return literally the Hamiltonian a_dag a a + a_dag a_dag a for the use in any Hamiltonian that uses more than
    just a resonator or Duffing part. A more general type of quantum element on a physical chip can have this type of interaction.
    One example is a three wave mixing element used in signal amplification called a Superconducting non-linear asymmetric inductive eLement
    (SNAIL in short). The code is a simple modification of the Duffing function and written in the same style.
    """
    a_dag = a.T.conj()
    n = np.matmul(a_dag, a)
    return np.matmul(n, a) + np.matmul(a_dag, n)


@hamiltonian_reg_deco
def int_XX(anhs):
    """
    Dipole type coupling.

    Parameters
    ----------
    anhs : Tensor list
        Annihilators.

    Returns
    -------
    Tensor
        coupling

    """
    a = anhs[0]
    b = anhs[1]
    a_dag = a.T.conj()
    b_dag = b.T.conj()
    return np.matmul(a_dag + a, b_dag + b)


@hamiltonian_reg_deco
def int_YY(anhs):
    """
    Dipole type coupling.

    Parameters
    ----------
    anhs : Tensor list
        Annihilators.

    Returns
    -------
    Tensor
        coupling

    """
    a = anhs[0]
    b = anhs[1]
    a_dag = a.T.conj()
    b_dag = b.T.conj()
    return -np.matmul(a_dag - a, b_dag - b)


@hamiltonian_reg_deco
def x_drive(a):
    """
    Semiclassical drive.

    Parameters
    ----------
    a : Tensor
        Annihilator.

    Returns
    -------
    Tensor
        Number operator.

    """
    a_dag = a.T.conj()
    return a_dag + a


@hamiltonian_reg_deco
def y_drive(a):
    """
    Semiclassical drive.

    Parameters
    ----------
    a : Tensor
        Annihilator.

    Returns
    -------
    Tensor
        Number operator.

    """
    a_dag = a.T.conj()
    return 1.0j * (a_dag - a)


@hamiltonian_reg_deco
def z_drive(a):
    """
    Semiclassical drive.

    Parameters
    ----------
    a : Tensor
        Annihilator.

    Returns
    -------
    Tensor
        Number operator.

    """
    a_dag = a.T.conj()
    return np.matmul(a_dag, a)
